import pandas as pd
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.graphics.tsaplots import plot_pacf, plot_acf
from matplotlib import pyplot
from base_model import base_model
import sys

if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")


class sarima(base_model):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._model = None
    
    def build_model(self, data):
        print('Building sarima model.')

        size = int(len(data) * self._config['train_size'])
        data = data[self._config['label_column']].values
        train, test = data[:size].tolist(), data[size:]
        
        model = SARIMAX(train,
                        order=(self._config['trend_order'][0], self._config['trend_order'][1], self._config['trend_order'][2]),
                        seasonal_order=(self._config['season_order'][0], self._config['season_order'][1], self._config['season_order'][2], self._config['season_order'][3])
                        )
        fitted_model = model.fit()
        
        predictions = fitted_model.forecast(steps = len(test))

        print('Evaluating arima performance.')
        self.evaluation(test, predictions)

        model = SARIMAX(train,
                      order=(self._config['trend_order'][0], self._config['trend_order'][1], self._config['trend_order'][2]),
                      seasonal_order=(self._config['season_order'][0], self._config['season_order'][1], self._config['season_order'][2], self._config['season_order'][3])
                      )
        model_fitted = model.fit()
        self._model = model_fitted
        return model
    
    def predict(self, X_new):
        return self._model.forecast(steps = step)

    def predict_next_n(self, steps):
        return self._model.get_forecast(steps=steps).predicted_mean

    def autocorrelation(self, data, lags = 20):
        print("Autocorrelation:")
        try:
            plot_acf(data[self._config['label_column']], lags = lags)
            pyplot.show()
        except:
            print('Data time step is below 20, please specify paramter lags to be below 20.')
        return None
    
    def partial_autocorrelation(self, data, lags = 20):
        print("Partial Autocorrelation:")
        try:
            plot_pacf(data[self._config['label_column']], lags = lags)
            pyplot.show()
        except:
            print('Data time step is below 20, please specify paramter lags to be below 20.')
        return None
    
    def residual_plot(self):
        print("Residual Plot:")
        df = pd.DataFrame(self._model.resid)
        df.plot()
        df.plot(kind='kde')
        pyplot.show()
        print("residual mean is {}".format(sum(self._model.resid)/len(self._model.resid)))
        return None
    
    def residual_density_plot(self):
        print("Residual Density Plot:")
        pd.DataFrame(self._model.resid).plot(kind='kde')
        pyplot.show()
        return None        
        
    
if __name__ == "__main__":
    import pandas as pd
    data = pd.read_csv("../../../feature_mart.csv")
    data = data[-500:]
    
    user_config = {
                   'label_column' : 'USEP',
                   'train_size' : 0.95,
                   'seed' : 33,
                   'trend_order' : (1, 1, 1),
                   'season_order' : (1, 0, 1, 48)
                   }
    sarima_shell = sarima(user_config)

##    sarima_shell.autocorrelation(data)
##    sarima_shell.partial_autocorrelation(data)
    sarima_shell.build_model(data)
    print(sarima_shell.predict_next_n(10))
    



from sklearn.ensemble import RandomForestRegressor
from .base_model import base_model


class random_forest(base_model):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._model = None
    
    def build_model(self, data):
        print('Building random forest model.')
        model = RandomForestRegressor(n_estimators = self._config['n_estimators'], 
                                      bootstrap = self._config['bootstrap'],
                                      criterion = self._config['criterion'],
                                      max_features = self._config['max_features'])
        self._model = model

        X_train, X_test, y_train, y_test = self.split(data,
                                                      self._config['label_column'],
                                                      self._config['train_size'],
                                                      self._config['seed'])
        self._model.fit(X_train, y_train)
        y_predicted = self._model.predict(X_test)
        
        print('Evaluating random forest performance.')
        self.evaluation(y_test, y_predicted)
        return model
    
    def predict(self, X_new):
        return self._model.predict(X_new)

    
if __name__ == "__main__":
    import pandas as pd
    data = pd.read_csv("../../development/Feature_Mart/feature_mart_merged_1.csv")
    data = data.drop(['date','demand_clusters','moonrise','moonset','sunrise','sunset'], axis = 1)
    random_forest_shell = random_forest({
                                    'label_column' : 'USEP',
                                    'train_size' : 0.8,
                                    'seed' : 33,
                                    'n_estimators' : 5,
                                    'bootstrap' : True,
                                    'criterion' : 'mse',
                                    'max_features' : 'sqrt'})
    
    random_forest_shell.build_model(data)



from statsmodels.tsa.arima_model import ARIMA
from pandas.plotting import autocorrelation_plot
from matplotlib import pyplot
from statsmodels.graphics.tsaplots import plot_pacf
from base_model import base_model
import pandas as pd
import sys

if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")


class arima(base_model):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._model = None
    
    def build_model(self, data):
        print('Building arima model.')

        size = int(len(data) * self._config['train_size'])
        data = data[self._config['label_column']].values
        train, test = data[:size].tolist(), data[size:]

        model = ARIMA(train,
                        order=(self._config['lag'], self._config['differencing'], self._config['window_size']))
        fitted_model = model.fit()
        
        predictions = fitted_model.forecast(steps = len(test))[0]
        
        print('Evaluating arima performance.')
        self.evaluation(test, predictions)

        model = ARIMA(data,
                      order=(self._config['lag'], self._config['differencing'], self._config['window_size']))
        model_fitted = model.fit()
        self._model = model_fitted
        return model
    
    def predict(self, X_new):
        return self._model.predict(X_new)
    
    def predict_next_n(self, step):
        return self._model.forecast(steps = step)[0]

    def autocorrelation(self, data, number_of_time_step = 20):
        print("Autocorrelation")
        try:
            autocorrelation_plot(data[self._config['label_column']][:number_of_time_step])
            pyplot.show()
        except:
            print('Data time step is below 20, please specify paramter number of time step. to be below 20.')
        return None
    
    def partial_autocorrelation(self, data, lags = 20):
        print("Partial Autocorrelation")
        try:
            plot_pacf(data[self._config['label_column']], lags = lags)
            pyplot.show()
        except:
            print('Data time step is below 20, please specify paramter lags to be below 20.')
        return None
    
    def residual_plot(self):
        print("Residual Plot")
        df = pd.DataFrame(self._model.resid)
        df.plot()
        df.plot(kind='kde')
        pyplot.show()
        print("residual mean is {}".format(sum(self._model.resid)/len(self._model.resid)))
        return None
    
    def residual_density_plot(self):
        print("Residual Density Plot")
        pd.DataFrame(self._model.resid).plot(kind='kde')
        pyplot.show()
        return None        
        
    
if __name__ == "__main__":
    import pandas as pd
    data = pd.read_csv("../../../feature_mart.csv")
    data = data[-100:]
    
    user_config = {'lag' : 2,
                   'differencing' : 0,
                   'window_size' : 2,
                   'label_column' : 'USEP',
                   'train_size' : 0.8,
                   'seed' : 33}
    arima_shell = arima(user_config)
##    arima_shell.autocorrelation(data)
##    arima_shell.partial_autocorrelation(data)
    arima_shell.build_model(data)
    print(arima_shell.predict_next_n(10))
##    arima_shell.residual_plot()
    



import xgboost as xgb
from base_model import base_model


class xgboost(base_model):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._model = None
    
    def build_model(self, data):
        print('Building xgboost model.')
        model = xgb.XGBRegressor(learning_rate = self._config['learning_rate'], 
                                      n_estimators = self._config['n_estimators'],
                                      max_depth = self._config['max_depth'],
                                      min_child_weight = self._config['min_child_weight'],
                                      gamma = self._config['gamma'],
                                      subsample=self._config['subsample'],
                                      colsample_bytree=self._config['colsample_bytree'],
                                      objective=self._config['objective'],
                                      scale_pos_weight=self._config['scale_pos_weight'])
        self._model = model

        X_train, X_test, y_train, y_test = self.split(data,
                                                      self._config['label_column'],
                                                      self._config['train_size'],
                                                      self._config['seed'])
        self._model.fit(X_train, y_train)
        y_predicted = self._model.predict(X_test)


        print('Evaluating xgboost performance.')
        self.evaluation(y_test, y_predicted)
        return model
    
    def predict(self, X_new):
        return self._model.predict(X_new)

    
if __name__ == "__main__":
    import pandas as pd
    data = pd.read_csv("../../../feature_mart.csv")
    data = data[-100:]
    xgboost_shell = xgboost({    
                'label_column': 'USEP',
                'train_size' : 0.8,
                'seed' : 33,
                'learning_rate': 0.1,
                'n_estimators': 150,
                'max_depth': 20,
                'min_child_weight': 4,
                'gamma': 0,
                'subsample': 0.7,
                'colsample_bytree': 0.8,
                'objective': 'reg:linear',
                'scale_pos_weight': 1,
                'seed': 1234})
    
    xgboost_shell.build_model(data)



from wwo_hist import retrieve_hist_data
import pandas as pd

class weather:
    def __init__(self, config):
        self.config = config

    def get_weather_data(self):
        retrieve_hist_data(self.config['api_key'],
                           self.config['location_list'],
                           self.config['start_date'],
                           self.config['end_date'],
                           self.config['frequency'],
                           location_label = self.config['location_label'],
                           export_csv = True,
                           store_df = False)
        return None
    def proprogate_half_hourly(self, data):
        data = data.drop_duplicates()
        # parse datetime columns
        data['date_time'] = pd.to_datetime(data['date_time'])
        # extract period and date
        data['period'] = data['date_time'].apply(lambda x:x.hour)
        data['date'] = data['date_time'].apply(lambda x:x.date())
        # parsing
        data2 = data.copy()
        data['period'] = data['period'] * 2
        data2['period'] = (data2['period'] * 2) + 1
        full = pd.concat([data,data2]).sort_values(by = ['date', 'period'])
        full["period"] = full["period"] + 1
        return full
    


if __name__ == '__main__':
    user_config = {'frequency' : 1,
                   'start_date' : '01-Jan-2010',
                   'end_date' : '31-Jan-2010',
                   'api_key' : '2c9e967a17ba475087893244201503',
                   'location_list' : ['singapore'],
                   'location_label' : False}
    
    weather_tool = weather(user_config)
    weather_tool.get_weather_data()
    data = pd.read_csv('singapore.csv')
    data = weather_tool.proprogate_half_hourly(data)
    print(len(data))



from keras.models import Sequential
from keras.layers import LSTM
from keras.layers import Dense
from .base_model import base_model
import numpy as np
import sys

if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")

class lstm_univariate(base_model):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._model = None

    def split_sequence(self, sequence, n_steps):
        X, y = [], []
        for i in range(len(sequence)):
            end_ix = i + n_steps
            if end_ix > len(sequence)-1:
                break
            seq_x, seq_y = sequence[i:end_ix], sequence[end_ix]
            X.append(seq_x)
            y.append(seq_y)
        return np.array(X), np.array(y)
    
    def build_model(self, data):
        print('Building LSTM model.')
        X, y = self.split_sequence(data[self._config['label_column']], self._config['n_steps'])
        size = int(len(data) * self._config['train_size'])
        X_train = X[:size]
        y_train = y[:size]
        X_test = X[size:]
        y_test = y[size:]

        X_train = X_train.reshape((X_train.shape[0], X_train.shape[1], 1))
        X_test = X_test.reshape((X_test.shape[0], X_test.shape[1], 1))

        
        model = Sequential()
        model.add(LSTM(self._config['n_neuron'],
                       activation = self._config['activation'],
                       input_shape = (self._config['n_steps'], 1)
                       )
                  )
        model.add(Dense(1))
        
        model.compile(optimizer=self._config['optimizer'],
                      loss=self._config['loss'])
        
        model.fit(X_train,
                  y_train,
                  epochs=self._config['n_epochs'],
                  verbose=1)
        
        self._model = model

        predict = []
        for x in X_test:
            x = x.reshape((1, self._config['n_steps'], 1))
            yhat = model.predict(x, verbose=0)
            predict.append(yhat[0])
        
        print('Evaluating LSTM performance.')
        self.evaluation(y_test, predict)
        return model
    
    def predict(self, X_new):
        return self._model.predict(X_new)

    
if __name__ == "__main__":
    import pandas as pd
    data = pd.read_csv("../../feature_mart.csv")
    user_config = {'label_column' : 'USEP',
                   'n_steps' : 7,
                   'train_size' : 0.8,
                   'n_neuron' : 5,
                   'activation' : 'relu',
                   'optimizer' : 'adam',
                   'loss' : 'mse',
                   'n_epochs' : 2}
    
    LSTM_shell = lstm_univariate(user_config)
    LSTM_shell.build_model(data)



## reference: https://github.com/datasets/oil-prices
## adapt to get only daily data
import datetime

from dataflows import Flow, PackageWrapper, ResourceWrapper, validate, filter_rows
from dataflows import add_metadata, dump_to_path, load, set_type, printer

def rename_resources(package: PackageWrapper):
    package.pkg.descriptor['resources'][0]['name'] = 'brent-daily'
    package.pkg.descriptor['resources'][0]['path'] = 'brent-daily.csv'
    package.pkg.descriptor['resources'][1]['name'] = 'wti-daily'
    package.pkg.descriptor['resources'][1]['path'] = 'wti-daily.csv'

    yield package.pkg
    res_iter = iter(package)
    for res in  res_iter:
        yield res.it
    yield from package


def filter_out_empty_rows(rows):
    for row in rows:
        if row['Date']:
            yield row


oil_prices = Flow(
    add_metadata(
        name="oil-prices",
        title= "Brent and WTI Spot Prices",
        descriptor="A variety of temporal granularities for Europe Brent and WTI (West Texas Intermediate) Spot Prices.",
        sources=[
            {
                "name": "Daily Europe Brent Spot Price",
                "path": "https://www.eia.gov/dnav/pet/hist_xls/RBRTEd.xls",
                "title": "Daily Europe Brent Spot Price"
            },
            {
                "name": "Daily Cushing, OK WTI Spot Price",
                "path": "http://www.eia.gov/dnav/pet/hist_xls/RWTCd.xls",
                "title": "Daily Cushing, OK WTI Spot Price"
            }
        ],
        licenses=[
            {
                "name": "ODC-PDDL-1.0",
                "path": "http://opendatacommons.org/licenses/pddl/",
                "title": "Open Data Commons Public Domain Dedication and License v1.0"
            }
        ],
        keywords=["Oil","Brent","WTI","Oil Prices","eia","oil eia"],
        views=[
            {
                "name": "graph",
                "title": "Europe Brent Spot Price FOB (Dollars per Barrel)",
                "resourceName": "brent-day",
                "specType": "simple",
                "spec": {
                "type": "line",
                    "group": "Date",
                    "series": ["Brent Spot Price"]
                }
            }
        ]
    ),
    load(
        load_source='https://www.eia.gov/dnav/pet/hist_xls/RBRTEd.xls',
        format='xls',
        sheet=2,
        skip_rows=[1,2,3],
        headers=['Date', 'Price']
    ),
    load(
        load_source='http://www.eia.gov/dnav/pet/hist_xls/RWTCd.xls',
        format='xls',
        sheet=2,
        skip_rows=[1,2,3],
        headers=['Date', 'Price']
    ),
    rename_resources,
    set_type('Date', resources=None, type='date', format='any'),
    validate(),
    filter_out_empty_rows,
    dump_to_path(),
)


if __name__ == '__main__':
    oil_prices.process()



import pandas as pd

class feature_engineering:
    def __init__(self):
        print('successfully substantiate feature engineering class.')

    def concat_columns(self, dataframe_list):
        return pd.concat(dataframe_list, axis=1)

    def concat_rows(self, dataframe_list):
        return pd.concat(dataframe_list)

    def join(self, dataframe_list, on = None, how = None):
        main_df = dataframe_list[0]
        for df in dataframe_list[1:]:
            main_df = main_df.merge(df, on = on, how = how)
        return main_df
    
    def parse_datetime(self, data, datetime_col, remove = True):
        try:
            data[datetime_col] = pd.to_datetime(data[datetime_col])
        except:
            print('Not date column')
        try:
            data['year'] = data[datetime_col].apply(lambda x: x.year)
        except:
            print('No information on YEAR')
        try:
            data['month'] = data[datetime_col].apply(lambda x:x.month)
        except:
            print('No information on MONTH')
        try:
            data['day'] = data[datetime_col].apply(lambda x:x.day)
        except:
            print('No information on DAY')
        if remove:
            return data.drop(datetime_col, axis = 1)
        else:
            return data
        
    def get_dummies(self, data, column_list):
        for col in column_list:
            data = self.concat_columns([data.drop(col, axis=1), pd.get_dummies(data[col])])
        return data

    def get_weekday(self, data):
        pass
    def get_peak_offpeak_shoulder():
        pass
    
if __name__ == '__main__':
    import pandas as pd
    demand = pd.read_csv('../development/Feature_Mart/demand_feature_mart.csv')
    weather = pd.read_csv('../development/Feature_Mart/weather_feature_mart.csv')
    FE = feature_engineering()
    df = FE.join([demand, weather], on = ['date', 'period'], how ='left')
    df = FE.parse_datetime(df, 'date', remove = True)
    df = FE.get_dummies(df, 'demand_clusters')
    print(df['off-peak'])
    print(df.columns)
                





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nextstep",
    version="0.0.14",
    author="Yang Yuesong",
    author_email="yangyuesongyys@gmail.com",
    description="USEP price prediction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/YangYuesong0323/nextstep",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires = [
        'wwo-hist == 0.0.4',
        'tensorflow == 2.1.0',
        'statsmodels == 0.11.0',
        'dataflows == 0.0.71',
        'Keras == 2.3.1',
        'pandas == 1.0.1',
        'numpy == 1.18.1',
        'scikit-learn == 0.22.1']
)



from sklearn.ensemble import AdaBoostRegressor
from .base_model import base_model


class adaboost(base_model):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._model = None
    
    def build_model(self, data):
        print('Building adaboost model.')
        model = AdaBoostRegressor(base_estimator = self._config['base_estimator'],
                                  n_estimators = self._config['n_estimators'],
                                  learning_rate = self._config['learning_rate'],
                                  loss = self._config['loss'],
                                  random_state = self._config['seed'])
        self._model = model
        
        X_train, X_test, y_train, y_test = self.split(data,
                                                      self._config['label_column'],
                                                      self._config['train_size'],
                                                      self._config['seed'])
        
        self._model.fit(X_train, y_train)
        y_predicted = self._model.predict(X_test)

        print('Evaluating adaboost performance.')
        self.evaluation(y_test, y_predicted)
        return model
    
    def predict(self, X_new):
        return self._model.predict(X_new)
    
if __name__ == "__main__":
    import pandas as pd
    data = pd.read_csv("../../development/Feature_Mart/feature_mart_merged_1.csv")
    data = data.drop(['date','demand_clusters','moonrise','moonset','sunrise','sunset'], axis = 1)
    from random_forest import random_forest
    random_forest_shell = random_forest({
                                    'label_column' : 'USEP',
                                    'train_size' : 0.8,
                                    'seed' : 33,
                                    'n_estimators' : 5,
                                    'bootstrap' : True,
                                    'criterion' : 'mse',
                                    'max_features' : 'sqrt'})
    
    rf = random_forest_shell.build_model(data)
    
    user_config = {'label_column' : 'USEP',
                   'train_size' : 0.9,
                   'seed' : 33,
                   'base_estimator': rf,
                   'n_estimators' : 10,
                   'learning_rate' : 1,
                   'loss' : 'square'}
    adaboost_shell = adaboost(user_config)
    adaboost_shell.build_model(data)
    



from sklearn.model_selection import train_test_split
from sklearn import metrics
import numpy as np

class base_model:
    def __init__(self):
       print('Welcome to EMC Challenge')
    def split(self, data, label_column, train_size, seed):
        y = data[label_column]
        X = data.drop(label_column, axis = 1)
        X_train,X_test,y_train,y_test = train_test_split(X, y, train_size=train_size, random_state=seed)
        return X_train, X_test, y_train, y_test

    def split_noshuffle(self, data, label_column, train_size, seed):
        y = data[label_column]
        X = data.drop(label_column, axis = 1)
        X_train,X_test,y_train,y_test = train_test_split(X, y, train_size=train_size, random_state=seed, shuffle=False)
        return X_train, X_test, y_train, y_test
    
    def evaluation(self, y_pre, y_true):
        print("MAE : {}".format(metrics.mean_absolute_error(y_true, y_pre)))
        print("MSE : {}".format(metrics.mean_squared_error(y_true, y_pre)))
        print("RMSE : {}".format(np.sqrt(metrics.mean_squared_error(y_true, y_pre))))
        return None
